part of dartcord;

/// The heart of a Discord bot.
class Client {
  String _token;
  ClientOptions _options;
  DateTime _startTime;
  _RestClient _http;
  _SocketManager _ws;
  _EventController _events;

  /// The logged in user.
  ClientUser user;

  /// The bot's OAuth2 app.
  ClientOAuth2Application app;

  /// All of the guilds the bot is in.
  Map<String, Guild> guilds;

  /// All of the channels the bot is in.
  Map<String, dynamic> channels;

  /// All of the users the bot can see. Does not always have offline users
  /// without forceFetchUsers enabled.
  Map<String, User> users;

  /// Whether or not the client is ready.
  bool ready = false;

  /// The current version.
  String version = _Constants.clientVersion;

  /// The client's internal shards.
  Map<int, _DiscordSocket> shards;

  /// Emitted when the client is ready.
  Stream<ReadyEvent> onReady;

  /// Emitted when a message is received.
  Stream<MessageCreateEvent> onMessageCreate;

  /// Emitted when a message is edited.
  Stream<MessageUpdateEvent> onMessageUpdate;

  /// Emitted when a message is deleted.
  Stream<MessageDeleteEvent> onMessageDelete;

  /// Emitted when a channel is created.
  Stream<ChannelCreateEvent> onChannelCreate;

  /// Emitted when a channel is updated.
  Stream<ChannelUpdateEvent> onChannelUpdate;

  /// Emitted when a channel is deleted.
  Stream<ChannelDeleteEvent> onChannelDelete;

  /// Emitted when a member is banned.
  Stream<GuildBanAddEvent> onGuildBanAdd;

  /// Emitted when a user is unbanned.
  Stream<GuildBanRemoveEvent> onGuildBanRemove;

  /// Emitted when the client joins a guild.
  Stream<GuildCreateEvent> onGuildCreate;

  /// Emitted when a guild is updated..
  Stream<GuildUpdateEvent> onGuildUpdate;

  /// Emitted when the client leaves a guild.
  Stream<GuildDeleteEvent> onGuildDelete;

  /// Emitted when a guild becomes unavailable.
  Stream<GuildUnavailableEvent> onGuildUnavailable;

  /// Emitted when a member joins a guild.
  Stream<GuildMemberAddEvent> onGuildMemberAdd;

  /// Emitted when a member is updated.
  Stream<GuildMemberUpdateEvent> onGuildMemberUpdate;

  /// Emitted when a user leaves a guild.
  Stream<GuildMemberRemoveEvent> onGuildMemberRemove;

  /// Emitted when a member's presence is changed.
  Stream<PresenceUpdateEvent> onPresenceUpdate;

  /// Emitted when a user starts typing.
  Stream<TypingEvent> onTyping;

  /// Emitted when a role is created.
  Stream<RoleCreateEvent> onRoleCreate;

  /// Emitted when a role is updated.
  Stream<RoleUpdateEvent> onRoleUpdate;

  /// Emitted when a role is deleted.
  Stream<RoleDeleteEvent> onRoleDelete;

  /// Creates and logs in a new client.
  Client(this._token, [this._options]) {
    DiscordLogging.setup();
    if (this._options == null) {
      this._options = new ClientOptions();
    }
    w_transport.globalTransportPlatform = vmTransportPlatform;

    this.guilds = new Map<String, Guild>();
    this.channels = new Map<String, dynamic>();
    this.users = new Map<String, User>();
    this.shards = new Map<int, _DiscordSocket>();

    this._http = new _RestClient(this);
    this._events = new _EventController(this);
    this._ws = new _SocketManager(this);
  }

  /// The client's uptime.
  Duration get uptime => new DateTime.now().difference(_startTime);

  /// Destroys the WebSocket connection, SS connection or server, and all streams.
  Future<Null> destroy() async {
    await this._ws.close();
    await this._events.destroy();
    return null;
  }

  /// Starts the [Client] and related WebSocket.
  Future start() async {
    return await this._ws.createAll();
  }

  /// Fetches a [User] object.
  ///
  /// Throws an [Exception] if the HTTP request errored.
  ///     Client.fetchUser("user id");
  Future<User> fetchUser(dynamic user) async {
    final String id = Util.resolve('user', user);

    final w_transport.Response r = await this._http.send('GET', _DiscordEndpoints.user(id));
    return new User._new(this, r.body.asJson() as Map<String, dynamic>);
  }

  /// Fetches an [Invite] object.
  ///
  /// Throws an [Exception] if the HTTP request errored.
  ///     Client.getInvite("invite code");
  Future<Invite> fetchInvite(String code) async {
    final w_transport.Response r =
        await this._http.send('GET', _DiscordEndpoints.invite(code));
    return new Invite._new(this, r.body.asJson() as Map<String, dynamic>);
  }

  /// Fetches an [OAuth2Info] object.
  ///
  /// Throws an [Exception] if the HTTP request errored
  ///     Client.getOAuth2Info("app id");
  Future<OAuth2Info> fetchOAuth2Info(dynamic app) async {
    final String id = Util.resolve('app', app);

    final w_transport.Response r = await this
        ._http
        .send('GET', _OAuth2Endpoints.authorize(id));
    return new OAuth2Info._new(this, r.body.asJson() as Map<String, dynamic>);
  }

  /// Creates a message.
  ///
  /// Throws an [Exception] if the HTTP request errored.
  ///     Channel.createMessage("My content!");
  Future<Message> createMessage(String channelID, String content,
      {bool tts = false, EmbedBuilder builder, String nonce, bool disableEveryone}) async {
    String newContent;
    if (disableEveryone == true ||
        (disableEveryone == null && this._options.disableEveryone)) {
      newContent = content
          .replaceAll("@everyone", "@\u200Beveryone")
          .replaceAll("@here", "@\u200Bhere");
    } else {
      newContent = content;
    }
    final w_transport.Response r = await this._http.send(
        'POST', _DiscordEndpoints.channelMessages(channelID), body: <String, dynamic>{
      "content": newContent,
      "tts": tts,
      "nonce": nonce,
      "embed": (builder != null ? builder.build() : null)
    });
    return new Message._new(
        this, r.body.asJson() as Map<String, dynamic>);
  }
}
