part of dartcord;

/// An abstract class that indicates that something can be built.
abstract class Buildable {
  /// Builds the class.
  Map<String, dynamic> build();
}

/// A helper class to create sendable embeds.
class EmbedBuilder implements Buildable {
  /// The title of the embed.
  String title;
  /// The type of embed.
  String type = 'rich';
  /// The description of the embed.
  String description;
  /// The URL of the embed.
  String url;
  // The timestamp of the embed.
  //DateTime timestamp;
  /// The color code of the embed.
  int color;
  /// The footer of the embed.
  EmbedFooterBuilder footer = new EmbedFooterBuilder();
  /// The image of the embed.
  EmbedImageBuilder image = new EmbedImageBuilder();
  /// The thumbnail of the embed.
  EmbedThumbnailBuilder thumbnail = new EmbedThumbnailBuilder();
  /// The video of the embed.
  EmbedVideoBuilder video = new EmbedVideoBuilder();
  /// The provider of the embed.
  EmbedProviderBuilder provider = new EmbedProviderBuilder();
  /// The author of the embed.
  EmbedAuthorBuilder author = new EmbedAuthorBuilder();
  /// A list containing all fields in the embed.
  List<EmbedFieldBuilder> fields = new List<EmbedFieldBuilder>();

  /// Creates an EmbedBuilder.
  EmbedBuilder();

  /// Creates an embed from data.
  EmbedBuilder._fromData(Map<String, dynamic> data) {
    title = data['title'];
    type = data['type'];
    description = data['description'];
    url = data['url'];
    color = data['color'];
  }

  /// Adds a field to the embed fields list.
  void addField(String title, String value, bool inline) {
    var f = new EmbedFieldBuilder();
    f.name = title;
    f.value = value;
    f.inline = inline;
    fields.add(f);
  }

  /// Compiles the [EmbedBuilder] into a valid embed.
  @override
  Map<String, dynamic> build() {
    return <String, dynamic> {
      'title': this.title,
      'type': this.type,
      'description': this.description,
      'url': this.url,
      'color': this.color,
      'footer': (footer != null ? footer.build() : null),
      'image': (image != null ? image.build() : null),
      'thumbnail': (thumbnail != null ? thumbnail.build() : null),
      'video': (video != null ? video.build() : null),
      'provider': (provider != null ? provider.build() : null),
      'author': (author != null ? author.build() : null),
      'fields': (fields != null ? new List.from(fields.map((val) => val.build())) : null)
    };
  }
}

/// A builder class for embed footers.
class EmbedFooterBuilder implements Buildable {
  /// The text in the footer.
  String text;
  /// The icon URL for the footer.
  String iconUrl;
  /// A proxied icon URL for the footer.
  String proxyIconUrl;

  @override
  Map<String, dynamic> build() {
    return <String, dynamic> {
      'text': text,
      'icon_url': iconUrl,
      'proxy_icon_url': proxyIconUrl
    };
  }
  /// Creates a new embed footer builder.
  EmbedFooterBuilder([this.text, this.iconUrl, this.proxyIconUrl]);
}

/// A builder class for embed thumbnails.
class EmbedThumbnailBuilder implements Buildable {
  /// The thumbnail URL.
  String url;
  /// A proxied thumbnail URL.
  String proxyUrl;
  /// The height of the thumbnail.
  int height;
  /// The width of the thumbnail.
  int width;
  @override
  Map<String, dynamic> build() {
    return <String, dynamic> {
      'url': url,
      'proxy_url': proxyUrl,
      'height': height,
      'width': width
    };
  }

  /// Creates a new embed thumbnail builder.
  EmbedThumbnailBuilder([this.url, this.proxyUrl, this.height, this.width]);
}

/// A builder class for embed videos.
class EmbedVideoBuilder implements Buildable {
  /// The URL of the video.
  String url;
  /// The height of the video.
  int height;
  /// The width of the video.
  int width;
  @override
  Map<String, dynamic> build() {
    return <String, dynamic> {
    'url': url,
    'height': height,
    'width': width
    };
  }

  /// Creates a new embed video builder.
  EmbedVideoBuilder([this.url, this.height, this.width]);
}

/// A builder class for embed images.
class EmbedImageBuilder implements Buildable {
  /// The name of the image.
  String name;
  /// The URL of the image.
  String url;
  @override
  Map<String, dynamic> build() {
    return <String, dynamic>{
      'name': name,
      'url': url
    };
  }

  /// Creates a new embed image builder.
  EmbedImageBuilder([this.name, this.url]);
}

/// A builder class for embed providers.
class EmbedProviderBuilder implements Buildable {
  /// The name of the provider.
  String name;
  /// The URL of the provider.
  String url;
  @override
  Map<String, dynamic> build() {
    return <String, dynamic>{
      'name': name,
      'url': url
    };
  }

  /// Creates a new embed provider builder.
  EmbedProviderBuilder([this.name, this.url]);
}

/// A builder class for embed authors.
class EmbedAuthorBuilder implements Buildable {
  /// The name of the author.
  String name;
  /// The URL of the author.
  String url;
  /// The icon URL of the author.
  String iconUrl;
  /// A proxied icon URL of the author;
  String proxyIconUrl;
  @override
  Map<String, dynamic> build() {
    return <String, dynamic>{
      'name': name,
      'url': url,
      'icon_url': iconUrl,
      'proxy_icon_url': proxyIconUrl
    };
  }

  /// Creates a new embed author builder.
  EmbedAuthorBuilder([this.name, this.url, this.iconUrl, this.proxyIconUrl]);
}

/// A builder class for embed fields.
class EmbedFieldBuilder implements Buildable{
  /// The name of the field.
  String name;
  /// The value of the field.
  String value;
  /// Whether the field appears inline.
  bool inline = false;
  @override
  Map<String, dynamic> build() {
    return <String, dynamic>{
      'name': (name != null ? name : ''),
      'value': (value != null ? value : ''),
      'inline': inline
    };
  }

  /// Creates a new embed field builder.
  EmbedFieldBuilder([this.name, this.value, this.inline]);
}


/// Encodes a timestamp into JSON.
dynamic encodeTimestamp(dynamic item) {
  if(item is DateTime) {
    return item.toIso8601String();
  }
  return item;
}