import 'dart:async';
import 'dart:io';

import 'package:dartcord/dartcord.dart' as discord;

void _fail(String message) {
  _failNoExit(message);
  exit(1);
}

void _failNoExit(String message) {
  discord.DiscordLogging.log(discord.LoggingLevel.critical, 'Tests failed: ' + message);
}

void _success() {
  discord.DiscordLogging.log(discord.LoggingLevel.info, 'Tests succeeded.');
  exit(0);
}

Future main() async {
  var cfg = new Map<String, String>();
  var env = Platform.environment;
  if (env['DISCORD_TOKEN'] == null || env['TESTING_CHAN_ID'] == null) {
    _fail('Environment variables DISCORD_TOKEN and TESTING_CHAN_ID have not been set. Please refer to the repository.');
  } else {
    cfg['DISCORD_TOKEN'] = env['DISCORD_TOKEN'];
    cfg['TESTING_CHAN_ID'] = env['TESTING_CHAN_ID'];
  }
  var bot = new discord.Client(env['DISCORD_TOKEN'], new discord.ClientOptions(disableEveryone: true, disabledEvents: [ 'TYPING_START' ]));
  discord.DiscordLogging.info('Starting tests. This may take a while.');

  if (env['TRAVIS_BRANCH'] == null) cfg['TRAVIS_BRANCH'] = 'local';
  if (env['TRAVIS_COMMIT'] == null) cfg['TRAVIS_COMMIT'] = 'local';
  if (env['TRAVIS_BUILD_NUMBER'] == null) cfg['TRAVIS_BUILD_NUMBER'] = 'local';
  else {
    cfg['TRAVIS_BRANCH'] = env['TRAVIS_BRANCH'];
    cfg['TRAVIS_COMMIT'] = env['TRAVIS_COMMIT'];
    cfg['TRAVIS_BUILD_NUMBER'] = env['TRAVIS_BUILD_NUMBER'];
  }

  new Timer(const Duration(seconds: 60), () {
    _fail('Timed out waiting for messages.');
  });

  var testingChannelId = env['TESTING_CHAN_ID'];
  const endTest = '--end';

  bot.onMessageCreate.listen((discord.MessageCreateEvent e) async {
    try {
      var m = e.message;

      if (m.channel.id == testingChannelId &&
          m.author.id == bot.user.id &&
          m.content == endTest) {
        await m.delete();
        await m.channel.createMessage("Tests completed successfully!");
        await bot.destroy();
        _success();
      }
    } catch (e) {
      _failNoExit('Error thrown during testing:');
      throw e;
    }
  });

  bot.onReady.listen((e) async {
      var channel = bot.channels[testingChannelId];
      if (env['TRAVIS_BUILD_NUMBER'] == 'local') {
        await channel.createMessage("Started local development test.");
      } else {
        await channel.createMessage(
            "Started Travis CI test with build number `#${cfg['TRAVIS_BUILD_NUMBER']}` from commit `${cfg['TRAVIS_COMMIT']}` on branch `${cfg['TRAVIS_BRANCH']}`");
      }
      var m = await channel.createMessage("Message test.");
      await m.edit("Edit test.");
      await m.delete();
      await channel.createMessage("Creating, editing and deleting messages tests complete.");
      var embed = new discord.EmbedBuilder();
      var auditLogs = await channel.guild.fetchAuditLogs() as discord.GuildAuditLogsContainer;
      int auditCount = 0;
      Map<int, String> audits = new Map<int, String>();
      for (var auditLog in auditLogs.auditLogs.values) {
        if (auditCount < 3) {
          audits[auditCount] = auditLog.actionType;
          auditCount++;
        }
      }
      await channel.createMessage('First three audit log action types: ' + audits.values.join(', '));
      await channel.createMessage("Audit log testing complete.");
      embed.author.name = "Discord.Dart CI Build";
      embed.description = "TRAVIS-CI test with build number ${cfg['TRAVIS_BUILD_NUMBER']} on branch ${cfg['TRAVIS_BRANCH']} started.";
      embed.addField('Result', 'Success', true);
      embed.addField('Build Number', cfg['TRAVIS_BUILD_NUMBER'], true);
      embed.addField('Branch', cfg['TRAVIS_BRANCH'], true);
      embed.footer.text = "Commit: ${cfg['TRAVIS_COMMIT']}";
      await channel.createMessage('e', builder: embed);
      await channel.createMessage("Embed test complete.");
      await channel.createMessage(endTest);
  });

  await bot.start();

}
